///////////////////////////////////////////////////////////////////////
// WMV plugin for VirtualDub 1.8.1+
// Copyright (C) 2007-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// IVDXVideoDecoder, IVDXVideoDecoderModel,
// and IVDXVideoSource implementations

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <vfw.h>


// Use VirtualDub's I/P decoder model?
//#define USE_DEFAULT_IP_MODEL

// Defer video decoding to VirtualDub?
//#define DEFER_DECODING


#include "vd2\plugin\vdinputdriver.h"
#include "Unknown.h"

#include "WinFile.h"
#include "DataVector.h"
#include "VideoCodec.h"

#include "ASFStreamNode.h"
#include "InputFileWMV.h"
#include "VideoSourceWMV.h"


#ifndef DEFER_DECODING

///////////////////////////////////////////////////////////////////////
//
//						VideoDecoderWMV
//
///////////////////////////////////////////////////////////////////////

class VideoDecoderWMV : public vdxunknown<IVDXVideoDecoder> {

private:
	InputFileWMV *const parentPtr;
	ASFStreamNode *const video;

	VideoCodec mCodec;

	sint64 mPrevFrame;
	DWORD dwActualFrame;

public:
	VideoDecoderWMV(InputFileWMV *const pp, ASFStreamNode *const asn);
	~VideoDecoderWMV();

	// IVDXVideoDecoder
	const void*         VDXAPIENTRY DecodeFrame(const void *inputBuffer, uint32 data_len, bool is_preroll, sint64 sampleNumber, sint64 targetFrame);
	uint32              VDXAPIENTRY GetDecodePadding();
	void                VDXAPIENTRY Reset();
	bool                VDXAPIENTRY IsFrameBufferValid();
	const VDXPixmap&    VDXAPIENTRY GetFrameBuffer();
	bool                VDXAPIENTRY SetTargetFormat(int format, bool useDIBAlignment);
	bool                VDXAPIENTRY SetDecompressedFormat(const VDXBITMAPINFOHEADER *pbih);
	bool                VDXAPIENTRY IsDecodable(sint64 sample_num);
	const void*         VDXAPIENTRY GetFrameBufferBase();
};


VideoDecoderWMV::VideoDecoderWMV(InputFileWMV *const pp, ASFStreamNode *const asn)
	: parentPtr (pp)
	, video     (asn)
{
	parentPtr->AddRef();

	if (video != NULL)
	{
		video->hCodec = NULL;
		dwActualFrame = 0xFFFFFFFF;
		mPrevFrame    = -1;
	}
	else
	{
		mPrevFrame = 0;
	}
}


VideoDecoderWMV::~VideoDecoderWMV()
{
	if (video != NULL) video->hCodec = NULL;
	parentPtr->Release();
}


const void *VDXAPIENTRY VideoDecoderWMV::DecodeFrame(
	const void *inputBuffer, uint32 data_len, bool is_preroll,
	sint64 streamFrame,	sint64 targetFrame)
{

//_RPT4(_CRT_WARN, "DecodeFrame %08X, %i, %I64i, %I64i\n", (int)inputBuffer,
//	(int)is_preroll, streamFrame, targetFrame);

	if (video != NULL)
	{
		DWORD dwFrame;
		
		if (streamFrame < 0) streamFrame = targetFrame;

		if ((uint64)streamFrame >= (uint64)video->dwPerceivedLength)
		{
			return mCodec.mFrameBuffer;
		}

		dwFrame = video->sample_map[streamFrame];

		// If dwFrame is the same as dwActualFrame,
		// then we've already decoded this frame.

		if (dwFrame != dwActualFrame)
		{
			DWORD dwFlags = video->sample_list[dwFrame].flags;

			dwActualFrame = dwFrame;

			if ((dwFlags & ASI_FLAG_KEYFRAME) == 0) {
				dwFlags = ICDECOMPRESS_NOTKEYFRAME;
			} else {
				dwFlags = 0;
			}

			if (is_preroll) {
				dwFlags |= ICDECOMPRESS_PREROLL;
			} else {
				mPrevFrame = streamFrame;
			}

			mCodec.Decompress((LPVOID)inputBuffer, data_len, dwFlags);
		}
	}

	return mCodec.mFrameBuffer;
}


uint32 VDXAPIENTRY VideoDecoderWMV::GetDecodePadding()
{
	return 0;
}


void VDXAPIENTRY VideoDecoderWMV::Reset()
{
	if (video != NULL)
	{
		mPrevFrame    = -1;
		dwActualFrame = 0xFFFFFFFF;
	}
}


bool VDXAPIENTRY VideoDecoderWMV::IsFrameBufferValid()
{
	return (mPrevFrame >= 0);
}


const VDXPixmap& VDXAPIENTRY VideoDecoderWMV::GetFrameBuffer()
{
	return mCodec.mPixmap;
}


bool VDXAPIENTRY VideoDecoderWMV::SetTargetFormat(int format, bool useDIBAlignment)
{
	if (video != NULL)
	{
		Reset();

		if (mCodec.Init((BITMAPINFOHEADER *)video->pFormat, format))
		{
			video->hCodec = mCodec.hCodec;
			return true;
		}

		if (mCodec.hCodec != NULL) return false;
		video->hCodec = NULL;

		// Missing codec
		mCodec.Inform("MISSING CODEC");

		return (format == mCodec.mPixmap.format
		     || format == nsVDXPixmap::kPixFormat_Null);
	}

	// Emulate
	if (mCodec.Init(NULL, format))
	{
		// No video stream
		mCodec.Inform("NO VIDEO");

		return (format == mCodec.mPixmap.format
		     || format == nsVDXPixmap::kPixFormat_Null);
	}

	return false;
}


bool VDXAPIENTRY VideoDecoderWMV::SetDecompressedFormat(const VDXBITMAPINFOHEADER *pbih)
{
	// In Fast Recompress mode, this gets
	// called with biCompression = 'WMV3'
	return false;
}


bool VDXAPIENTRY VideoDecoderWMV::IsDecodable(sint64 sample_num64)
{
	if (video != NULL)
	{
		if (sample_num64 >= 0 && sample_num64 < (sint64)video->dwPerceivedLength)
		{
			return (sample_num64 == (mPrevFrame + 1));
		}
		return false;
	}
	return true;
}


const void *VDXAPIENTRY VideoDecoderWMV::GetFrameBufferBase()
{
	return mCodec.mFrameBuffer;
}


#endif	// DEFER_DECODING


#ifndef USE_DEFAULT_IP_MODEL

///////////////////////////////////////////////////////////////////////
//
//						VideoDecoderModelWMV
//
///////////////////////////////////////////////////////////////////////

class VideoDecoderModelWMV : public vdxunknown<IVDXVideoDecoderModel> {

public:
	VideoDecoderModelWMV(InputFileWMV *const pp, ASFStreamNode *const asn);
	~VideoDecoderModelWMV();

	// IVDXVideoDecoderModel
	void    VDXAPIENTRY Reset();
	void    VDXAPIENTRY SetDesiredFrame(sint64 frame_num);
	sint64  VDXAPIENTRY GetNextRequiredSample(bool& is_preroll);
	int     VDXAPIENTRY GetRequiredCount();

private:
	InputFileWMV*   const parentPtr;
	ASFStreamNode*  const video;

	sint64  mLastFrame;
	sint64  mTargetFrame;
	sint64  mNextFrame;

	bool IsKey(sint64 sample);
};


VideoDecoderModelWMV::VideoDecoderModelWMV(InputFileWMV *const pp, ASFStreamNode *const asn)
	: parentPtr (pp)
	, video     (asn)
{
	parentPtr->AddRef();
	Reset();
}


VideoDecoderModelWMV::~VideoDecoderModelWMV()
{
	parentPtr->Release();
}


void VDXAPIENTRY VideoDecoderModelWMV::Reset()
{
	mLastFrame = -1;
}


bool VideoDecoderModelWMV::IsKey(sint64 sample)
{
	if (sample >= 0 && sample < (sint64)video->dwPerceivedLength)
	{
		DWORD dwActual = video->sample_map[sample];
		if (sample == 0 || video->sample_map[sample - 1] != dwActual)
		{
			return ((video->sample_list[dwActual].flags & ASI_FLAG_KEYFRAME) != 0);
		}
	}
	return false;
}


void VDXAPIENTRY VideoDecoderModelWMV::SetDesiredFrame(sint64 frame_num)
{
	mTargetFrame = frame_num;

	// Do we already have the frame?
	if (mLastFrame == frame_num)
	{
		mNextFrame = -1;	// Flag this for GetNextRequiredSample
		return;
	}

	mNextFrame = frame_num;

	// If it's a Delta frame, locate a previous Key
	if (!IsKey(frame_num))
	{
		while (mNextFrame > 0)
		{
			--mNextFrame;
			if (mNextFrame == mLastFrame)
			{
				// Start from the frame following that one
				++mNextFrame;
				break;
			}
			if (IsKey(mNextFrame)) break;
		}
	}
}


sint64 VDXAPIENTRY VideoDecoderModelWMV::GetNextRequiredSample(bool& is_preroll)
{
	sint64 frame;

	if (mNextFrame == mTargetFrame || mNextFrame < 0)
	{
		is_preroll = false;
		mLastFrame = mTargetFrame;
		frame      = mNextFrame;
		mNextFrame = -1;
	}
	else
	{
		is_preroll = true;
		frame      = mNextFrame++;
	}

	return frame;
}


int VDXAPIENTRY VideoDecoderModelWMV::GetRequiredCount()
{
	if (mNextFrame < 0) return 0;

	return (int)((mTargetFrame - mNextFrame) + 1);
}

#endif	// USE_DEFAULT_IP_MODEL


VideoSourceWMV::VideoSourceWMV(InputFileWMV *const pp, ASFStreamNode *const vasn, ASFStreamNode *const aasn)
	: parentPtr (pp)
	, video     (vasn)
	, audio     (aasn)
{
	parentPtr->AddRef();

	errmode = kErrorModeDecodeAnyway;
	frame_buffer = NULL;
	dwPrevSample = 0xFFFFFFFF;

	if (video != NULL)
	{
		long max = 0;
		for (DWORD i = 0; i < video->dwActualLength; ++i)
		{
			if (video->sample_list[i].size > max)
			{
				max = video->sample_list[i].size;
			}
		}
		frame_buffer = new unsigned char[max];

		dwLength = video->dwPerceivedLength;
	}
	else
	{
		// Emulate a video source
		const WAVEFORMATEX *wfex = (WAVEFORMATEX *)audio->pFormat;

		dwLength = (DWORD)(wfex->nSamplesPerSec * audio->Duration * 0.001 + 0.5);
	}
}


VideoSourceWMV::~VideoSourceWMV()
{
	delete[] frame_buffer;
	parentPtr->Release();
}


int VDXAPIENTRY VideoSourceWMV::AddRef()
{
	return vdxunknown<IVDXStreamSource>::AddRef();
}


int VDXAPIENTRY VideoSourceWMV::Release()
{
	return vdxunknown<IVDXStreamSource>::Release();
}


void *VDXAPIENTRY VideoSourceWMV::AsInterface(uint32 iid)
{
	if (iid == IVDXVideoSource::kIID)
		return static_cast<IVDXVideoSource *>(this);

#ifdef EXPOSE_V3
	if (iid == IVDXStreamSourceV3::kIID)
		return static_cast<IVDXStreamSourceV3 *>(this);
#endif

	return vdxunknown<IVDXStreamSource>::AsInterface(iid);
}


void VDXAPIENTRY VideoSourceWMV::GetStreamSourceInfo(VDXStreamSourceInfo& srcInfo)
{
	if (video != NULL)
	{
		srcInfo.mSampleRate.mNumerator   = video->dwRate;
		srcInfo.mSampleRate.mDenominator = video->dwScale;
	}
	else
	{
		// Emulated video source
		const WAVEFORMATEX *wfex = (WAVEFORMATEX *)audio->pFormat;

		srcInfo.mSampleRate.mNumerator   = wfex->nSamplesPerSec;
		srcInfo.mSampleRate.mDenominator = 1000;
	}

	srcInfo.mSampleCount = dwLength;

	// TODO:
	srcInfo.mPixelAspectRatio.mNumerator   = 1;
	srcInfo.mPixelAspectRatio.mDenominator = 1;
}


#ifdef EXPOSE_V3

void VDXAPIENTRY VideoSourceWMV::GetStreamSourceInfoV3(VDXStreamSourceInfoV3& srcInfo)
{
	GetStreamSourceInfo(srcInfo.mInfo);

	srcInfo.mFlags      = 0;
	srcInfo.mfccHandler = 0;

	if (video != NULL)
	{
		srcInfo.mFlags = VDXStreamSourceInfoV3::kFlagVariableSizeSamples;
		if (video->pFormat != NULL && video->iFormatLen >= sizeof(BITMAPINFOHEADER))
		{
			srcInfo.mfccHandler = ((BITMAPINFOHEADER *)video->pFormat)->biCompression;
		}
	}
}

#endif


bool VDXAPIENTRY VideoSourceWMV::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead)
{
	DWORD dwAVCDCRSize;
	DWORD dwStart = (DWORD)lStart64;
	DWORD dwDup = -1;

	if (video == NULL)
	{
		// Emulated video source
		if (lBytesRead) *lBytesRead = 4096;
		if (lSamplesRead) *lSamplesRead = 1;
		return (lpBuffer == NULL || cbBuffer >= 4096);
	}

	// Reading out-of-bounds?
	if ((uint64)lStart64 >= (uint64)video->dwPerceivedLength)
	{
		if (lBytesRead) *lBytesRead = 0;
		if (lSamplesRead) *lSamplesRead = 0;
		return true;
	}

	dwStart = video->sample_map[lStart64];

	if (lStart64 > 0 && video->sample_map[lStart64 - 1] == dwStart)
	{
		// dwStart is a duplicate of the previous frame
		if (lBytesRead) *lBytesRead = 0;
		if (lSamplesRead) *lSamplesRead = 1;
		return true;
	}

	const uint32 framesize = video->sample_list[dwStart].size;

	dwAVCDCRSize = (video->sample_list[dwStart].flags & ASI_FLAG_KEYFRAME)?
		video->iAVCDCRLen: 0;

	if (lpBuffer == NULL || cbBuffer < (framesize + dwAVCDCRSize))
	{
		if (lBytesRead) *lBytesRead = framesize + dwAVCDCRSize;
		if (lSamplesRead) *lSamplesRead = 1;
		return (lpBuffer == NULL);
	}

//_RPT2(_CRT_WARN, "VideoSourceWMV::Read %i, %i\n", lStart, dwPrevSample);

	// read data, but only if it's new data
	if (dwStart != dwPrevSample)
	{
		dwPrevSample = dwStart;
		parentPtr->ReadData(frame_buffer, dwStart, 0, framesize, video);
	}

	if (dwAVCDCRSize) memcpy(lpBuffer, video->pAVCDCR, dwAVCDCRSize);
	memcpy((char *)lpBuffer + dwAVCDCRSize, frame_buffer, framesize);

	if (lBytesRead) *lBytesRead = framesize + dwAVCDCRSize;
	if (lSamplesRead) *lSamplesRead = 1;
	return true;
}


const void *VDXAPIENTRY VideoSourceWMV::GetDirectFormat()
{
	if (video != NULL) return video->pFormat;
	return NULL;
}


int VDXAPIENTRY VideoSourceWMV::GetDirectFormatLen()
{
	if (video != NULL) return video->iFormatLen;
	return 0;
}


IVDXStreamSource::ErrorMode VDXAPIENTRY VideoSourceWMV::GetDecodeErrorMode()
{
	return errmode;
}


void VDXAPIENTRY VideoSourceWMV::SetDecodeErrorMode(IVDXStreamSource::ErrorMode mode)
{
	errmode = mode;
}


bool VDXAPIENTRY VideoSourceWMV::IsDecodeErrorModeSupported(IVDXStreamSource::ErrorMode mode)
{
	// TODO
	return true;
}


bool VDXAPIENTRY VideoSourceWMV::IsVBR()
{
	return false;
}


sint64 VDXAPIENTRY VideoSourceWMV::TimeToPositionVBR(sint64 us)
{
	if (video != NULL)
	{
		return (sint64)(
			( (double)us * video->dwRate)
			/ (1000000.0 * video->dwScale) + 0.5
		);
	}
	else
	{
		// Emulated video source
		const WAVEFORMATEX *wfex =
			(WAVEFORMATEX *)audio->pFormat;

		return (sint64)(
			((double)us * wfex->nSamplesPerSec) / 1000000000.0 + 0.5
		);
	}
}


sint64 VDXAPIENTRY VideoSourceWMV::PositionToTimeVBR(sint64 samples)
{
	if (video != NULL)
	{
		return (sint64)(
			( (double)samples * video->dwScale * 1000000.0)
			/ (double)video->dwRate + 0.5
		);
	}
	else
	{
		// Emulated video source
		const WAVEFORMATEX *wfex =
			(WAVEFORMATEX *)audio->pFormat;

		return (sint64)(
			(double)samples / (1000.0 * wfex->nSamplesPerSec) + 0.5
		);
	}
}


void VDXAPIENTRY VideoSourceWMV::GetVideoSourceInfo(VDXVideoSourceInfo& info)
{
	if (video != NULL)
	{
		const BITMAPINFOHEADER *bmih = (BITMAPINFOHEADER *)video->pFormat;

		info.mWidth  = bmih->biWidth;
		info.mHeight = bmih->biHeight;
		info.mFlags  = 0;

#ifdef USE_DEFAULT_IP_MODEL
		info.mDecoderModel = VDXVideoSourceInfo::kDecoderModelDefaultIP;
#else
		info.mDecoderModel = VDXVideoSourceInfo::kDecoderModelCustom;
#endif
	}
	else
	{
		// Emulated video source
		info.mWidth  = 320;
		info.mHeight = 240;
		info.mFlags  = 0;

		info.mDecoderModel = VDXVideoSourceInfo::kDecoderModelDefaultIP;
	}
}


bool VDXAPIENTRY VideoSourceWMV::CreateVideoDecoderModel(IVDXVideoDecoderModel **ppModel)
{
#ifdef USE_DEFAULT_IP_MODEL
	*ppModel = NULL;
	return false;
#else
	*ppModel = new VideoDecoderModelWMV(parentPtr, video);
	if (*ppModel == NULL) return false;

	(*ppModel)->AddRef();
	return true;
#endif
}


bool VDXAPIENTRY VideoSourceWMV::CreateVideoDecoder(IVDXVideoDecoder **ppDecoder)
{
#ifdef DEFER_DECODING
	*ppDecoder = NULL;
	return false;
#else
	*ppDecoder = new VideoDecoderWMV(parentPtr, video);
	if (*ppDecoder == NULL) return false;

	(*ppDecoder)->AddRef();
	return true;
#endif
}


void VDXAPIENTRY VideoSourceWMV::GetSampleInfo(sint64 sample_num, VDXVideoFrameInfo& frameInfo)
{
	if (video != NULL)
	{
		if (sample_num >= 0 && sample_num < (sint64)video->dwPerceivedLength)
		{
			DWORD dwActual = video->sample_map[sample_num];

			const ASFSampleInfo& asi = video->sample_list[dwActual];

			if (sample_num > 0 && video->sample_map[sample_num - 1] == dwActual)
			{
				frameInfo.mTypeChar  = 'D';
				frameInfo.mFrameType = kVDXVFT_Null;
			}
			else
			{
				if (asi.flags & ASI_FLAG_KEYFRAME)
				{
					frameInfo.mTypeChar  = 'K';
					frameInfo.mFrameType = kVDXVFT_Independent;
				}
				else
				{
					frameInfo.mTypeChar  = ' ';
					frameInfo.mFrameType = kVDXVFT_Predicted;
				}
			}
			frameInfo.mBytePosition = asi.pos;
		}
		else
		{
			frameInfo.mTypeChar  = 'M';
			frameInfo.mFrameType = kVDXVFT_Null;
		}
	}
	else
	{
		// Emulated video source
		if (IsKey(sample_num))
		{
			frameInfo.mTypeChar  = 'K';
			frameInfo.mFrameType = kVDXVFT_Independent;
		}
		else
		{
			frameInfo.mTypeChar  = ' ';
			frameInfo.mFrameType = kVDXVFT_Predicted;
		}
		frameInfo.mBytePosition = -1;
	}
}


bool VDXAPIENTRY VideoSourceWMV::IsKey(sint64 sample_num)
{
	if (video != NULL)
	{
		if (sample_num >= 0 && sample_num < (sint64)video->dwPerceivedLength)
		{
			DWORD dwActual = video->sample_map[sample_num];
			if (sample_num == 0 || video->sample_map[sample_num - 1] != dwActual)
			{
				return ((video->sample_list[dwActual].flags & ASI_FLAG_KEYFRAME) != 0);
			}
		}
		return false;
	}
	else
	{
		// Emulated video source
		const WAVEFORMATEX *wfex = (WAVEFORMATEX *)audio->pFormat;

		return (((sample_num * 100) % wfex->nSamplesPerSec) == 0);
	}
}


sint64 VDXAPIENTRY VideoSourceWMV::GetFrameNumberForSample(sint64 sample_num)
{
	return sample_num;
}


sint64 VDXAPIENTRY VideoSourceWMV::GetSampleNumberForFrame(sint64 frame_num)
{
	return frame_num;
}


sint64 VDXAPIENTRY VideoSourceWMV::GetRealFrame(sint64 frame_num)
{
	if (video != NULL)
	{
		if (frame_num >= 0 && frame_num < (sint64)video->dwPerceivedLength)
		{
			// If this is a dropped frame, seek backward
			// and return the actual frame it duplicates:

			DWORD dwActual = video->sample_map[frame_num];

			while (frame_num > 0)
			{
				--frame_num;
				if (video->sample_map[frame_num] != dwActual)
				{
					++frame_num;
					break;
				}
			}
		}
	}

	return frame_num;
}


sint64 VDXAPIENTRY VideoSourceWMV::GetSampleBytePosition(sint64 sample_num)
{
	if (video != NULL)
	{
		if (sample_num >= 0 && sample_num < (sint64)video->dwPerceivedLength)
		{
			DWORD dwActual = video->sample_map[sample_num];
			return video->sample_list[dwActual].pos;
		}
	}
	return -1;
}

